program running_cplot
	use sep

	implicit none

	integer, external	:: sreed, srite
	integer		:: i, j, k, w, k1, kwind, zpos
	integer		:: n1, nt
	real		:: o1, ot
	real		:: d1, dt
	real		:: fpeak, nwave
	real		:: mint, maxt
	integer		:: minti, maxti
	integer		:: ntwind, nwind

	real,		allocatable, dimension (:,:)	:: f1, f2
	complex,	allocatable, dimension (:)		:: cplot

	call initpar()

	call from_aux ('f1', "n1", n1)
	call from_aux ('f1', "o1", o1);
	call from_aux ('f1', "d1", d1);
	call from_aux ('f1', "n2", nt);
	call from_aux ('f1', "o2", ot);
	call from_aux ('f1', "d2", dt);

	call from_param ("fpeak", fpeak, 10.)
	call from_param ("nwave", nwave, 1.)
	call from_param ("zpos", zpos, 1)
	call from_param ("mint", mint, ot)
	call from_param ("maxt", maxt, ot + (nt-1)*dt)

	minti = mint / dt + 1
	maxti = maxt / dt + 1

	nwind = maxti - minti + 1 ! number of time windows equal number of time samples

	! determine size of time window by the time period of the dominant frequency
	ntwind = nwave / fpeak/sqrt(2.) / dt
	ntwind = ntwind / 2 * 2 + 1

	! allocate for time window
	allocate (cplot (ntwind))
	allocate (f1 (n1, nt))
	allocate (f2 (n1, nt))

	! write crossplot file's history details
	call to_history ("n1", ntwind)
	call to_history ("d1", 1.)
	call to_history ("o1", 0.)
	call to_history ("n2", 1)
	call to_history ("d2", 0.)
	call to_history ("o2", 0.)
	call to_history ("n3", nwind)
	call to_history ("d3", dt)
	call to_history ("o3", ot)
	call to_history ("esize", 8)

	! read x,z displacement files
	if (sreed ("f1", f1, n1 * nt * sizeof (f1(1,1))) .eq. 0) call seperr ("error reading input file 1")
	if (sreed ("f2", f2, n1 * nt * sizeof (f2(1,1))) .eq. 0) call seperr ("error reading input file 2")

	! loop over time windows
	do w = minti, maxti

		cplot = cmplx (0., 0.)

		k1 = w ! k1 is considered at center of time window
		kwind = 1
		! sum displacement of field along time axis in this time window
		do k = k1 - ntwind / 2, k1 + ntwind / 2 ! loop from beginning to end of time window
			if (k < 1) then
				kwind = kwind  + 1
				cycle
			else if (k > nt) then
				cplot (kwind) = cplot (kwind - 1)
				kwind = kwind + 1
				cycle
			endif

			! put x displacements in real part and z displacement in imaginary part
			cplot (kwind) = cmplx (f1 (zpos, k), f2 (zpos, k))
			kwind = kwind  + 1
		end do

		! write out the crossplot of displacements	
		if (srite ("out", cplot, ntwind * sizeof (cplot(1))) .eq. 0) call seperr ("error writing to displacement crossplot file")

	end do	! end time window loop

	deallocate (f1, f2, cplot)

end program running_cplot
