module sg_io_type_mod


	type io_type
		logical :: snap_P, snap_S
		logical :: snap_vx, snap_vz
		logical :: snap_ux, snap_uz
		logical :: snap_rot

		logical :: snap_Pb, snap_Sb
		logical :: snap_vxb, snap_vzb
		logical :: snap_uxb, snap_uzb
		logical :: snap_rotb

		logical :: rec_P, rec_S
		logical :: rec_vx, rec_vz
		logical :: rec_ux, rec_uz
		logical :: rec_rot

		logical :: rec_Pb, rec_Sb
		logical :: rec_vxb, rec_vzb
		logical :: rec_uxb, rec_uzb
		logical :: rec_rotb

		logical	:: in_vx_data, in_vz_data, in_P_data, in_S_data, in_vx_src, in_vz_src

		integer    :: jsnap,jrec
	end type


end module

module sg_io_mod
  use sep
  use sg_distrib_type_mod
  use sg_io_type_mod
  use sg_model_type_mod

  implicit none

  real, pointer,private :: slice2d_write (:,:), slice2d_write_s(:,:), slice2d_read (:,:), slice2d_read_s(:,:)
  real, pointer,private :: recs(:) ! temporary contigous array for receiver data before writing it out to file.
  real, pointer,private :: srcs(:) ! temporary contigous array for virtual source data before writing it out to file.

contains

subroutine io_init(model_s,io_s)
    type(io_type)		:: io_s
    type(model_type)	:: model_s
	integer				:: nsrcs, nrecs

    call from_param("jrec",io_s%jrec,1)
    call from_param("jsnap",io_s%jsnap,10)

	nrecs = size (model_s%recs,2)
	nsrcs = size (model_s%srcs,2)

    call create_snap_file (model_s,io_s,"snap_P",io_s%snap_P)
    call create_snap_file (model_s,io_s,"snap_S",io_s%snap_S)
    call create_snap_file (model_s,io_s,"snap_vx",io_s%snap_vx)
    call create_snap_file (model_s,io_s,"snap_vz",io_s%snap_vz)
    call create_snap_file (model_s,io_s,"snap_ux",io_s%snap_ux)
    call create_snap_file (model_s,io_s,"snap_uz",io_s%snap_uz)
    call create_snap_file (model_s,io_s,"snap_rot",io_s%snap_rot)

    call create_snap_file (model_s,io_s,"snap_Pb",io_s%snap_Pb)
    call create_snap_file (model_s,io_s,"snap_Sb",io_s%snap_Sb)
    call create_snap_file (model_s,io_s,"snap_vxb",io_s%snap_vxb)
    call create_snap_file (model_s,io_s,"snap_vzb",io_s%snap_vzb)
    call create_snap_file (model_s,io_s,"snap_uxb",io_s%snap_uxb)
    call create_snap_file (model_s,io_s,"snap_uzb",io_s%snap_uzb)
    call create_snap_file (model_s,io_s,"snap_rotb",io_s%snap_rotb)

	! receiver files
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_P",io_s%rec_P, nrecs, "receiver")
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_S",io_s%rec_S, nrecs, "receiver")
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_vx",io_s%rec_vx, nrecs, "receiver")
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_vz",io_s%rec_vz, nrecs, "receiver")
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_ux",io_s%rec_ux, nrecs, "receiver")
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_uz",io_s%rec_uz, nrecs, "receiver")
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_rot",io_s%rec_rot, nrecs, "receiver")

	! source files
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_Pb",io_s%rec_Pb, nsrcs, "virtual source")
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_Sb",io_s%rec_Sb, nsrcs, "virtual source")
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_vxb",io_s%rec_vxb, nsrcs, "virtual source")
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_vzb",io_s%rec_vzb, nsrcs, "virtual source")
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_uxb",io_s%rec_uxb, nsrcs, "virtual source")
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_uzb",io_s%rec_uzb, nsrcs, "virtual source")
    call create_rec_file (model_s%nt,model_s%dt,io_s%jrec,	"rec_rotb",io_s%rec_rotb, nrecs, "receiver")

	call create_input_file ("in_vx_data", io_s%in_vx_data)
	call create_input_file ("in_vz_data", io_s%in_vz_data)
	call create_input_file ("in_P_data", io_s%in_P_data)
	call create_input_file ("in_S_data", io_s%in_S_data)
	call create_input_file ("in_vx_src", io_s%in_vx_src)
	call create_input_file ("in_vz_src", io_s%in_vz_src)

    allocate (slice2d_write_s (model_s%nx, model_s%nz))
    allocate (slice2d_write   (model_s%nx - model_s%nsponge_x_l - model_s%nsponge_x_u, model_s%nz - model_s%nsponge_z_l - model_s%nsponge_z_u))
    allocate (slice2d_read_s  (model_s%nx, model_s%nz))
    allocate (slice2d_read    (model_s%nx - model_s%nsponge_x_l - model_s%nsponge_x_u, model_s%nz - model_s%nsponge_z_l - model_s%nsponge_z_u))
    allocate (recs (size (model_s%recs, 2)))
    allocate (srcs (size (model_s%srcs, 2)))
end subroutine
 
!##################
! subroutines for creating files for wavefield snaps, receivers, sources, input
!##################
subroutine create_snap_file(model_s,io_s,file,snap_do)
    type( model_type)  :: model_s 
    type(io_type)      :: io_s
    logical            :: snap_do
    character(len=*)   :: file
    character(len=256) :: temp
    integer ::n_t

    call from_param(file,temp,"none")
    snap_do=.false.
    n_t=ceiling(real(model_s%nt)/real(io_s%jsnap))
    if(temp(1:4)/="none") then
      snap_do=.true.
      call sep_put_data_axis_par (file, 1, model_s%nx - model_s%nsponge_x_l - model_s%nsponge_x_u, 0., model_s%dx, "x")
      call sep_put_data_axis_par (file, 2, model_s%nz - model_s%nsponge_z_l - model_s%nsponge_z_u, 0., model_s%dz, "z")
      call sep_put_data_axis_par (file, 3, n_t,													   0., model_s%dt*io_s%jsnap, "time")
    end if
end subroutine

subroutine create_rec_file(nt, dt, jrec, file, rec_do, nlocs, label)
	integer, intent(in)	:: nt, jrec, nlocs
	real, intent(in)	:: dt
    logical				:: rec_do
    character(len=*)	:: file, label
    character(len=256)	:: temp
    integer				:: n_t

    call from_param (file, temp, "none")
    rec_do = .false.
    n_t = ceiling (real (nt) / real (jrec))
    if (temp (1:4) /= "none") then
      rec_do = .true.
      call sep_put_data_axis_par (file, 1, nlocs, 0., 1., label)
      call sep_put_data_axis_par (file, 2, n_t, 0., dt*jrec, "time")
    end if
end subroutine

subroutine create_input_file (file,in_do)
    logical            :: in_do
    character(len=*)   :: file
    character(len=256) :: temp

    call from_param(file,temp,"none")
    in_do=.false.
    if(temp(1:4)/="none") then
      in_do=.true.
    end if
end subroutine

!##################
! subroutines for writing out files for wavefield snaps, receivers, sources
!##################
subroutine write_out_files(it,m_s,d_s,io_s,adj)
    type(model_type)	:: m_s
    type(distrib_type)	:: d_s
    type(io_type)		:: io_s
    integer				:: it
	logical				:: adj


	if (io_s%jsnap == 1 .or. mod (it, io_s%jsnap) == 1) then
		if (adj) then ! backward time direction
			if (io_s%snap_Pb)   call write_snap ("snap_Pb",  io_s%snap_Pb,  m_s, d_s, P_i)
			if (io_s%snap_Sb)   call write_snap ("snap_Sb",  io_s%snap_Sb,  m_s, d_s, S_i)
			if (io_s%snap_vxb)  call write_snap ("snap_vxb", io_s%snap_vxb, m_s, d_s, vx_i)
			if (io_s%snap_vzb)  call write_snap ("snap_vzb", io_s%snap_vzb, m_s, d_s, vz_i)
			if (io_s%snap_uxb)  call write_snap ("snap_uxb", io_s%snap_uxb, m_s, d_s, ux_i)
			if (io_s%snap_uzb)  call write_snap ("snap_uzb", io_s%snap_uzb, m_s, d_s, uz_i)
			if (io_s%snap_rotb)  call write_snap ("snap_rotb", io_s%snap_rotb, m_s, d_s, rot_i)
		else ! forward time direction
			if (io_s%snap_P)   call write_snap ("snap_P",  io_s%snap_P,  m_s, d_s, P_i)
			if (io_s%snap_S)   call write_snap ("snap_S",  io_s%snap_S,  m_s, d_s, S_i)
			if (io_s%snap_vx)  call write_snap ("snap_vx", io_s%snap_vx, m_s, d_s, vx_i)
			if (io_s%snap_vz)  call write_snap ("snap_vz", io_s%snap_vz, m_s, d_s, vz_i)
			if (io_s%snap_ux)  call write_snap ("snap_ux", io_s%snap_ux, m_s, d_s, ux_i)
			if (io_s%snap_uz)  call write_snap ("snap_uz", io_s%snap_uz, m_s, d_s, uz_i)
			if (io_s%snap_rot)  call write_snap ("snap_rot", io_s%snap_rot, m_s, d_s, rot_i)
		end if
    end if
    

    if(io_s%jrec == 1 .or. mod (it, io_s%jrec) == 1) then
		if (adj) then ! backward time direction
			if (io_s%rec_Pb)   call write_srcs ("rec_Pb",   io_s%rec_Pb,   it, m_s, d_s, iarray=P_i)
			if (io_s%rec_Sb)   call write_srcs ("rec_Sb",   io_s%rec_Sb,   it, m_s, d_s, iarray=S_i)
			if (io_s%rec_vxb)  call write_srcs ("rec_vxb",  io_s%rec_vxb,  it, m_s, d_s, iarray=vx_i)
			if (io_s%rec_vzb)  call write_srcs ("rec_vzb",  io_s%rec_vzb,  it, m_s, d_s, iarray=vz_i)
			if (io_s%rec_uxb)  call write_srcs ("rec_uxb",  io_s%rec_uxb,  it, m_s, d_s, iarray=ux_i)
			if (io_s%rec_uzb)  call write_srcs ("rec_uzb",  io_s%rec_uzb,  it, m_s, d_s, iarray=uz_i)
			if (io_s%rec_rotb) call write_srcs ("rec_rotb", io_s%rec_rotb, it, m_s, d_s, iarray=rot_i)
		else ! forward time direction
			if (io_s%rec_P)   call write_recs ("rec_P",   io_s%rec_P,   it, m_s, d_s, iarray=P_i)
			if (io_s%rec_S)   call write_recs ("rec_S",   io_s%rec_S,   it, m_s, d_s, iarray=S_i)
			if (io_s%rec_vx)  call write_recs ("rec_vx",  io_s%rec_vx,  it, m_s, d_s, iarray=vx_i)
			if (io_s%rec_vz)  call write_recs ("rec_vz",  io_s%rec_vz,  it, m_s, d_s, iarray=vz_i)
			if (io_s%rec_ux)  call write_recs ("rec_ux",  io_s%rec_ux,  it, m_s, d_s, iarray=ux_i)
			if (io_s%rec_uz)  call write_recs ("rec_uz",  io_s%rec_uz,  it, m_s, d_s, iarray=uz_i)
			if (io_s%rec_rot) call write_recs ("rec_rot", io_s%rec_rot, it, m_s, d_s, iarray=rot_i)
		endif
    end if
end subroutine

! save all receiver data into file
subroutine save_recs (tag, in_do, m_s, d_s, iarray)
	character(len=*)	:: tag
	logical				:: in_do
	type(model_type)	:: m_s
	type(distrib_type)	:: d_s
	integer				:: iarray, nrecs
	integer,external	:: srite

	if (.not. in_do) then
		call seperr ("io_mod: no name specified for receiver output file")
	else
		nrecs = size (m_s%recs,2)

		if (4 * nrecs * m_s%nt /= srite (tag, d_s%recdata (1 + nrecs * m_s%nt * iarray : nrecs * m_s%nt * (iarray + 1)), nrecs * m_s%nt * 4)) then
			write (0, *), "tag:", trim(tag)
			call seperr ("io_mod: error writing out recs")
		end if
	endif
end subroutine

! save all virtual source data into file
subroutine save_srcs (tag, in_do, m_s, d_s, iarray)
	character(len=*)	:: tag
	logical				:: in_do
	type(model_type)	:: m_s
	type(distrib_type)	:: d_s
	integer				:: iarray, nsrcs
	integer,external	:: srite

	if (.not. in_do) then
		call seperr ("io_mod: no name specified for virtual source output file")
	else
		nsrcs = size (m_s%srcs,2)

		if (4 * nsrcs * m_s%nt /= srite (tag, d_s%srcmodel (1 + nsrcs * m_s%nt * iarray : nsrcs * m_s%nt * (iarray + 1)), nsrcs * m_s%nt * 4)) then
			write (0, *), "tag:", trim(tag)
			call seperr ("io_mod: error writing out srcs")
		end if
	endif
end subroutine

! record wavefield at receiver locations
subroutine write_recs (tag, out_do, it, m_s, d_s, iarray, iarray_out)
	character(len=*)	:: tag
	logical				:: out_do
	type(model_type)	:: m_s
	type(distrib_type)	:: d_s
	integer				:: i, it, nrecs
	integer,external	:: srite, sreed
	integer				:: iarray, iloc (2), irec
	integer, optional	:: iarray_out

	do i = 1, size (recs)
		iloc = m_s%recs (:, i)
		recs(i) = d_s%all (iloc (1), iloc (2), iarray)
	end do

	if (present (iarray_out)) then
		nrecs = size (recs)

		d_s%recdata (1 + nrecs * m_s%nt * iarray_out + (it-1) * nrecs : nrecs * m_s%nt * iarray_out + it * nrecs) = recs (:)
	endif

	if (out_do) then
		if (4 * size (recs) /= srite (tag, recs, size(recs) * 4)) then
			write (0, *) trim (tag)
			call seperr ("error writing out recs")
		end if
	end if

end subroutine

! record wavefield at virtual source locations
subroutine write_srcs (tag, out_do, it, m_s, d_s, iarray, iarray_out)
	character(len=*)	:: tag
	logical				:: out_do
	type(model_type)	:: m_s
	type(distrib_type)	:: d_s
	integer				:: i, it, nsrcs
	integer,external	:: srite, sreed
	integer				:: iarray, iloc (2), isrc
	integer, optional	:: iarray_out


	do i = 1, size (srcs)
		iloc = m_s%srcs (:, i)
		srcs(i) = d_s%all (iloc (1), iloc (2), iarray)
	end do

	if (present (iarray_out)) then
		nsrcs = size (srcs)

		d_s%srcmodel (1 + nsrcs * m_s%nt * iarray_out + (it-1) * nsrcs : nsrcs * m_s%nt * iarray_out + it * nsrcs) = srcs (:)
	endif

	if (out_do) then
		if (4 * size (srcs) /= srite (tag, srcs, size(srcs) * 4)) then
			write (0, *) trim (tag)
			call seperr ("error writing out srcs")
		end if
	end if

end subroutine

! record wavefield at all locations
subroutine write_snap (tag, out_do, m_s, d_s, iarray)
	character (len=*)   :: tag
	logical				:: out_do
	type (model_type)   :: m_s
	type (distrib_type) :: d_s
	integer             :: ix, iz, ierr, it, jsnap
	integer,external    :: srite, sreed
	integer             :: iarray, ibz, iez, ibx, iex
	
	
	slice2d_write = 0.
	slice2d_write_s = 0.

	ibz = m_s %fz
	iez = m_s %fz + m_s %nz - 1
	ibx = m_s %fx
	iex = m_s %fx + m_s %nx - 1


	slice2d_write_s (ibx:iex, ibz:iez) = d_s %all (:, :, iarray)

	do iz = 1, size (slice2d_write, 2)
		do ix = 1, size (slice2d_write, 1)
			slice2d_write (ix, iz) = slice2d_write_s (ix + m_s%nsponge_x_l, iz + m_s%nsponge_z_l)
		end do 
	end do 

	if (out_do) then
		if (4 * size (slice2d_write) /= srite (tag, slice2d_write, size (slice2d_write) * 4)) then
			write (0,*), "=slice tag=", trim (tag)
			call seperr ("error writing out slice")
		end if
	end if
end subroutine

!##################
! subroutines for reversing the time axis of the virtual sources and the receivers
!##################

! reverse time axis of receiver data
subroutine reverse_recs (m_s, d_s, iarray)
	type(model_type)	:: m_s
	type(distrib_type)	:: d_s
	integer				:: iarray, nrecs, irec, it
	real 				:: temp

	nrecs = size (m_s%recs,2)

	do it = 1, floor (m_s%nt / 2.);
		do irec = 1, nrecs
			temp = d_s%recdata (nrecs * m_s%nt * iarray + nrecs * (it-1) + irec)
			d_s%recdata (nrecs * m_s%nt * iarray + nrecs * (it-1) + irec) = d_s%recdata (nrecs * m_s%nt * iarray + nrecs * (m_s%nt-it) + irec)
			d_s%recdata (nrecs * m_s%nt * iarray + nrecs * (m_s%nt-it) + irec) = temp
		end do
	end do
end subroutine

! reverse time axis of virtual sources
subroutine reverse_srcs (m_s, d_s, iarray)
	type(model_type)	:: m_s
	type(distrib_type)	:: d_s
	integer				:: iarray, nsrcs, isrc, it
	real 				:: temp

	nsrcs = size (m_s%srcs,2)

	do it = 1, floor (m_s%nt / 2.);
		do isrc = 1, nsrcs
			temp = d_s%srcmodel (nsrcs * m_s%nt * iarray + nsrcs * (it-1) + isrc)
			d_s%srcmodel (nsrcs * m_s%nt * iarray + nsrcs * (it-1) + isrc) = d_s%srcmodel (nsrcs * m_s%nt * iarray + nsrcs * (m_s%nt-it) + isrc)
			d_s%srcmodel (nsrcs * m_s%nt * iarray + nsrcs * (m_s%nt-it) + isrc) = temp
		end do
	end do
end subroutine

!##################
! subroutines for reading in virtual sources and receivers
!##################

! load all receiver data into memory
subroutine load_recs (tag, in_do, m_s, d_s, iarray)
	character(len=*)	:: tag
	logical				:: in_do
	type(model_type)	:: m_s
	type(distrib_type)	:: d_s
	integer				:: iarray, nrecs, nread
	integer,external	:: sreed

	if (.not. in_do) then
		call seperr ("io_mod: no name specified for receiver input file")
	else

		nrecs = size (m_s%recs,2)

		nread = sreed (tag, d_s%recdata (1 + nrecs * m_s%nt * iarray : nrecs * m_s%nt * (iarray + 1)), nrecs * m_s%nt * 4)
		if (nread .eq. 0) then
			write (0, *), "tag:", trim(tag)
			call seperr ("io_mod: error reading in recs")
		end if
	endif
end subroutine

! load all virtual source data into memory
subroutine load_srcs (tag, in_do, m_s, d_s, iarray)
	character(len=*)	:: tag
	logical				:: in_do
	type(model_type)	:: m_s
	type(distrib_type)	:: d_s
	integer				:: iarray, nsrcs, nread
	integer,external	:: sreed

	if (.not. in_do) then
		call seperr ("io_mod: no name specified for virtual source input file")
	else
		nsrcs = size (m_s%srcs,2)

		nread = sreed (tag, d_s%srcmodel (1 + nsrcs * m_s%nt * iarray : nsrcs * m_s%nt * (iarray + 1)), nsrcs * m_s%nt * 4)
		if (nread .eq. 0) then
			write (0, *), "tag:", trim(tag)
			call seperr ("io_mod: error reading in srcs")
		end if
	endif
end subroutine

! add receiver data to wavefield
subroutine read_recs (tag, in_do, it, m_s, d_s, rec_scale, wavetype, iarray1, iarray2, iarray_in)
	character(len=*)	:: tag
	logical				:: in_do
	type(model_type)	:: m_s
	type(distrib_type)	:: d_s
	integer				:: i, it, nrecs, nread
	integer,external	:: sreed
	integer				:: iloc (2), irec, wavetype
	integer				:: iarray1
	integer, optional	:: iarray2, iarray_in
	real				:: rec_scale

	if (in_do) then ! read from file
		nread = sreed (tag, recs, size(recs) * 4)
		if (nread .eq. 0) then
			write (0, *), "tag:", trim(tag)
			call seperr ("io_mod: error reading in recs")
		end if
	end if

	if (present (iarray_in)) then ! read from memory
		nrecs = size (recs)

		recs (:) = d_s%recdata (1 + nrecs * m_s%nt * iarray_in + (it-1) * nrecs : nrecs * m_s%nt * iarray_in + it * nrecs)
	endif

	do i = 1, size (recs)
		iloc = m_s%recs (:, i)

		if (present(iarray2)) then ! read the same data and split it into two arrays
			if (wavetype .eq. 1) then ! read pressure data into displacement fields
		
					d_s%all (iloc (1)+1, iloc (2),   iarray1) = d_s%all (iloc (1)+1, iloc (2)  , iarray1) + rec_scale * recs (i)
					d_s%all (iloc (1)-1, iloc (2),   iarray1) = d_s%all (iloc (1)-1, iloc (2)  , iarray1) - rec_scale * recs (i)
					d_s%all (iloc (1),   iloc (2)+1, iarray2) = d_s%all (iloc (1),   iloc (2)+1, iarray2) + rec_scale * recs (i)
					d_s%all (iloc (1),   iloc (2)-1, iarray2) = d_s%all (iloc (1),   iloc (2)-1, iarray2) - rec_scale * recs (i)
			else if (wavetype .eq. 2) then ! read shear data into displacement fields
		
					d_s%all (iloc (1)+1, iloc (2),   iarray2) = d_s%all (iloc (1)+1, iloc (2)  , iarray2) - rec_scale * recs (i)
					d_s%all (iloc (1)-1, iloc (2),   iarray2) = d_s%all (iloc (1)-1, iloc (2)  , iarray2) + rec_scale * recs (i)
					d_s%all (iloc (1),   iloc (2)+1, iarray1) = d_s%all (iloc (1),   iloc (2)+1, iarray1) + rec_scale * recs (i)
					d_s%all (iloc (1),   iloc (2)-1, iarray1) = d_s%all (iloc (1),   iloc (2)-1, iarray1) - rec_scale * recs (i)
	
	
			endif
		else ! read into just one array
				d_s%all (iloc (1), iloc (2), iarray1) = d_s%all (iloc (1), iloc (2), iarray1) + rec_scale * recs (i)
				!d_s%all (iloc (1), iloc (2), iarray1) = rec_scale * recs (i)
		endif
	end do
end subroutine

! add virtual source data to wavefield
subroutine read_srcs (tag, in_do, it, m_s, d_s, src_scale, timesign, iarray1, iarray_in)
	character(len=*)	:: tag
	logical				:: in_do
	type(model_type)	:: m_s
	type(distrib_type)	:: d_s
	integer				:: i, it, nsrcs, nread
	integer,external	:: sreed
	integer				:: iloc (2), isrc, timesign
	integer				:: iarray1
	integer, optional	:: iarray_in
	real				:: src_scale
	integer				:: stat

	if (in_do) then ! read from file
		nread = sreed (tag, srcs, size(srcs) * 4)
		if (nread .eq. 0) then
			write (0,*), "tag:", trim(tag)
			call seperr ("io_mod: error reading in srcs")
		end if
	endif

	if (present (iarray_in)) then ! read from memory
		nsrcs = size (srcs)

		srcs (:) = d_s%srcmodel (1 + nsrcs * m_s%nt * iarray_in + (it-1) * nsrcs : nsrcs * m_s%nt * iarray_in + it * nsrcs)
	endif

	do i = 1, size (srcs)
		iloc = m_s%srcs (:, i)

		d_s%all (iloc (1), iloc (2), iarray1) = d_s%all (iloc (1), iloc (2), iarray1) + src_scale * timesign * srcs (i)
		!d_s%all (iloc (1), iloc (2), iarray1) = src_scale * timesign * srcs (i)
	end do

end subroutine


subroutine create_arrays (d_s, m_s)
	type(distrib_type)	:: d_s
	type(model_type)	:: m_s
	real				:: dmin, dmax, wavel, vmin, vmax, min_vs, vp, vs, u14
	integer				:: iz, ix, i
	integer				:: nwavel_samples
	integer				:: nrec, nsrc

	allocate (d_s %rho       (m_s %nx, m_s %nz))
	allocate (d_s %lambda    (m_s %nx, m_s %nz))
	allocate (d_s %mu        (m_s %nx, m_s %nz))
	allocate (d_s %mu_avg    (m_s %nx, m_s %nz))

	call read_expand_medium ("rho",		m_s, d_s%rho)
	call read_expand_medium ("mu",		m_s, d_s%mu)
	call read_expand_medium ("lambda",	m_s, d_s%lambda)


    call from_param("min_vs", min_vs, 500.)

	vp = sqrt ((d_s %lambda (m_s%nx/2, m_s%nz/2) + 2 * d_s %mu (m_s%nx/2, m_s%nz/2)) * d_s %rho (m_s%nx/2, m_s%nz/2))
	vmin = vp
	vmax = vp
	do iz = 1, m_s%nz
		do ix = 1, m_s%nx
			vp = sqrt ((d_s %lambda (ix,iz) + 2 * d_s %mu (ix,iz)) * d_s %rho (ix,iz))
			vs = sqrt (d_s %mu (ix,iz) * d_s %rho (ix,iz))
			if (vp > 0.) then
				vmin = min (vp, vmin)
				vmax = max (vp, vmax)
			endif
			if (vs > min_vs) then
				vmin = min (vs, vmin)
				vmax = max (vs, vmax)
			else
				d_s %mu (ix, iz) = 0. ! set mu to zero if vs is too slow. This is to prevent instability.
			end if
		end do
	end do

	wavel = vmin / (m_s%fpeak*2.5)	! the 2.5 is because of the ricker source. It tapers down at about 2.5 times the central frequency.
	dmax = max (m_s%dx, m_s%dz)

	call from_param("nwavel_samples", nwavel_samples, 8)

	if (nwavel_samples*dmax > wavel) then	! at least nwavel_samples per shortest wavelength in field
		write (*,*) 'Reduce dxmax,dymax or dzmax to less than', wavel/nwavel_samples
		call seperr("Grid size too large for accuracy")
	endif

	dmin = min (m_s%dx, m_s%dz)
	if (m_s%dt .gt. 0.4949 * dmin/vmax) then
		write (*,*) 'Reduce dt to less than', 0.4949 * dmin/vmax
		call seperr ('Time Step too large for stability')
	endif


	! calculate harmonic average for mu at grid points of txz
	d_s%mu_avg = d_s%mu

	do iz = 1, m_s%nz-1
		do ix = 1, m_s%nx-1
			u14 = d_s%mu (ix, iz) * d_s%mu (ix+1, iz) * d_s%mu (ix, iz+1) * d_s%mu (ix+1, iz+1)

			if (abs (sum (d_s%mu (ix:ix+1, iz:iz+1)) - 4 * d_s%mu (ix, iz)) <.0001) then ! if this mu is near in value to its neighbours, use this d_s%mu
				d_s%mu_avg (ix, iz) = d_s%mu (ix, iz)
			else
				if (u14 .le. 1.e-4) then ! if this d_s%mu is different in value from the neighbouring mus, but the neighbourhood has small values, then use mu=0
					d_s%mu_avg (ix, iz) = 0.
				else ! otherwise, use harmonic average for this d_s%mu
					d_s%mu_avg (ix, iz) = (1. / d_s%mu (ix, iz)   + 1. / d_s%mu (ix+1, iz) +  &
										   1. / d_s%mu (ix, iz+1) + 1. / d_s%mu (ix+1, iz+1)) / 4.
					d_s%mu_avg (ix, iz) = 1. / d_s%mu_avg (ix, iz)
				endif
			end if
		end do
	end do

	! allocate wavefields
	allocate (d_s%all (m_s%nx, m_s%nz, 11))
	d_s%all = 0.

	d_s%P	=> d_s %all (:,:,P_i)
	d_s%S	=> d_s %all (:,:,S_i)
	d_s%vx	=> d_s %all (:,:,vx_i)
	d_s%vz	=> d_s %all (:,:,vz_i)
	d_s%ux	=> d_s %all (:,:,ux_i)
	d_s%uz	=> d_s %all (:,:,uz_i)
	d_s%txx	=> d_s %all (:,:,txx_i)
	d_s%tzz	=> d_s %all (:,:,tzz_i)
	d_s%txz	=> d_s %all (:,:,txz_i)
	d_s%rot	=> d_s %all (:,:,rot_i)
	d_s%rotc=> d_s %all (:,:,rotc_i)

	! allocate model
	nsrc = size (m_s%srcs, 2)
	allocate (d_s%srcmodel (nsrc * m_s%nt * 2))
	d_s%srcmodel = 0.

	d_s%vxs	=> d_s %srcmodel (1 + nsrc * m_s%nt * vxs_i : nsrc * m_s%nt * (vxs_i + 1))
	d_s%vzs	=> d_s %srcmodel (1 + nsrc * m_s%nt * vzs_i : nsrc * m_s%nt * (vzs_i + 1))

	! allocate data
	nrec = size (m_s%recs, 2)
	allocate (d_s%recdata (nrec * m_s%nt * 3))
	d_s%recdata = 0.

	d_s%vxr	=> d_s %recdata (1 + nrec * m_s%nt * vxr_i : nrec * m_s%nt * (vxr_i + 1))
	d_s%vzr	=> d_s %recdata (1 + nrec * m_s%nt * vzr_i : nrec * m_s%nt * (vzr_i + 1))
	d_s%Pr	=> d_s %recdata (1 + nrec * m_s%nt * Pr_i  : nrec * m_s%nt * (Pr_i  + 1))

end subroutine

subroutine read_expand_medium (tag, m_s, medium)
	! read medium parameters (lambda, mu, rho) and pad with the edge values
	character (len=*)   :: tag
	type (model_type)   :: m_s
	real, pointer, dimension (:,:) :: medium
	integer,external    :: sreed
	integer             :: bz, ez, bx, ex


	if (4 * size (slice2d_read) /= sreed (tag, slice2d_read, size (slice2d_read) * 4)) then
		write (0,*), "=slice tag=", trim(tag)
		call seperr("error reading in tag")
	end if

	call expand_slice (slice2d_read, slice2d_read_s)

	bz = m_s %fz
	ez = m_s %fz + m_s %nz - 1
	bx = m_s %fx
	ex = m_s %fx + m_s %nx - 1

	medium (:, :) = slice2d_read_s (bx:ex, bz:ez)

end subroutine 

subroutine expand_slice(small,big)
	real :: small (:,:), big(:,:)
	integer :: ix, iz, nsx, nsz, iloc

	nsx = (size (big, 1) - size (small, 1)) / 2
	nsz = (size (big, 2) - size (small, 2)) / 2

	do iz = 1, size (small, 2)
		do ix = 1, size (small, 1)
			big (ix+nsx, iz+nsz) = small (ix, iz)
		end do
	end do

	do iz = 1, size (big, 2)
		if (iz <= nsz) then
			iloc = nsz+1

		else if (iz > size (big,2) - nsz) then
				iloc = size (small,2)
		else
			iloc = iz - nsz
		end if

		do ix = 1, nsx
			big (ix, iz) = small (1, iloc)
			big (size (big, 1) - ix+1, iz) = small (size (small, 1), iloc)
		end do

	end do

	do iz = 1, nsz
		do ix = 1, size (big, 1)
			big (ix, iz) = big (ix, nsz+1)
			big (ix, size (big, 2) - iz + 1) = big (ix, size (big, 2) - nsz)
		end do
	end do
end subroutine
end module
